@@ -2,6 +2,7 @@
 namespace PhpAmqpLib\Channel;
 
 use PhpAmqpLib\Exception\AMQPBasicCancelException;
+use PhpAmqpLib\Exception\AMQPConnectionClosedException;
 use PhpAmqpLib\Exception\AMQPProtocolChannelException;
 use PhpAmqpLib\Exception\AMQPRuntimeException;
 use PhpAmqpLib\Message\AMQPMessage;
@@ -10,20 +11,23 @@ use PhpAmqpLib\Wire\AMQPWriter;
 
 class AMQPChannel extends AbstractChannel
 {
-    /** @var array */
+    /**
+     * @var callable[]
+     * @internal Use is_consuming() to check if there is active callbacks
+     */
     public $callbacks = array();
 
     /** @var bool Whether or not the channel has been "opened" */
     protected $is_open = false;
 
     /** @var int */
-    protected $default_ticket;
+    protected $default_ticket = 0;
 
     /** @var bool */
-    protected $active;
+    protected $active = true;
 
     /** @var array */
-    protected $alerts;
+    protected $alerts = array();
 
     /** @var bool */
     protected $auto_decode;
@@ -68,22 +72,29 @@ class AMQPChannel extends AbstractChannel
      * @see basic_publish()
      * @see publish_batch()
      */
-    private $publish_cache;
+    private $publish_cache = array();
 
     /**
      * Maximal size of $publish_cache
      *
      * @var int
      */
-    private $publish_cache_max_size;
+    private $publish_cache_max_size = 100;
+
+    /**
+     * Maximum time to wait for operations on this channel, in seconds.
+     * @var float $channel_rpc_timeout
+     */
+    private $channel_rpc_timeout;
 
     /**
      * @param \PhpAmqpLib\Connection\AbstractConnection $connection
      * @param null $channel_id
      * @param bool $auto_decode
+     * @param int $channel_rpc_timeout
      * @throws \Exception
      */
-    public function __construct($connection, $channel_id = null, $auto_decode = true)
+    public function __construct($connection, $channel_id = null, $auto_decode = true, $channel_rpc_timeout = 0)
     {
         if ($channel_id == null) {
             $channel_id = $connection->get_free_channel_id();
@@ -91,17 +102,10 @@ class AMQPChannel extends AbstractChannel
 
         parent::__construct($connection, $channel_id);
 
-        $this->publish_cache = array();
-        $this->publish_cache_max_size = 100;
-
         $this->debug->debug_msg('using channel_id: ' . $channel_id);
 
-        $this->default_ticket = 0;
-        $this->is_open = false;
-        $this->active = true; // Flow control
-        $this->alerts = array();
-        $this->callbacks = array();
         $this->auto_decode = $auto_decode;
+        $this->channel_rpc_timeout = $channel_rpc_timeout;
 
         try {
             $this->x_open();
@@ -111,6 +115,14 @@ class AMQPChannel extends AbstractChannel
         }
     }
 
+    /**
+     * @return bool
+     */
+    public function is_open()
+    {
+        return $this->is_open;
+    }
+
     /**
      * Tear down this object, after we've agreed to close with the server.
      */
@@ -121,6 +133,7 @@ class AMQPChannel extends AbstractChannel
         }
         $this->channel_id = $this->connection = null;
         $this->is_open = false;
+        $this->callbacks = array();
     }
 
     /**
@@ -148,10 +161,12 @@ class AMQPChannel extends AbstractChannel
      * @param int $reply_code
      * @param string $reply_text
      * @param array $method_sig
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function close($reply_code = 0, $reply_text = '', $method_sig = array(0, 0))
     {
+        $this->callbacks = array();
         if ($this->is_open === false || $this->connection === null) {
             $this->do_close();
 
@@ -164,11 +179,17 @@ class AMQPChannel extends AbstractChannel
             $method_sig[1]
         );
 
-        $this->send_method_frame(array($class_id, $method_id), $args);
+        try {
+            $this->send_method_frame(array($class_id, $method_id), $args);
+        } catch (\Exception $e) {
+            $this->do_close();
+
+            throw $e;
+        }
 
         return $this->wait(array(
             $this->waitHelper->get_wait('channel.close_ok')
-        ));
+        ), false, $this->channel_rpc_timeout );
     }
 
     /**
@@ -203,6 +224,7 @@ class AMQPChannel extends AbstractChannel
      * Enables/disables flow from peer
      *
      * @param $active
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function flow($active)
@@ -212,7 +234,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('channel.flow_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -244,6 +266,7 @@ class AMQPChannel extends AbstractChannel
 
     /**
      * @param string $out_of_band
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     protected function x_open($out_of_band = '')
@@ -257,7 +280,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('channel.open_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -279,6 +302,7 @@ class AMQPChannel extends AbstractChannel
      * @param bool $active
      * @param bool $write
      * @param bool $read
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function access_request(
@@ -302,7 +326,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('access.request_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -330,6 +354,7 @@ class AMQPChannel extends AbstractChannel
      * @param bool $nowait
      * @param array $arguments
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
     public function exchange_declare(
@@ -340,10 +365,9 @@ class AMQPChannel extends AbstractChannel
         $auto_delete = true,
         $internal = false,
         $nowait = false,
-        $arguments = null,
+        $arguments = array(),
         $ticket = null
     ) {
-        $arguments = $this->getArguments($arguments);
         $ticket = $this->getTicket($ticket);
 
         list($class_id, $method_id, $args) = $this->protocolWriter->exchangeDeclare(
@@ -366,7 +390,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('exchange.declare_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -384,6 +408,7 @@ class AMQPChannel extends AbstractChannel
      * @param bool $if_unused
      * @param bool $nowait
      * @param null $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
     public function exchange_delete(
@@ -408,7 +433,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('exchange.delete_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -429,6 +454,7 @@ class AMQPChannel extends AbstractChannel
      * @param bool $nowait
      * @param array $arguments
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
     public function exchange_bind(
@@ -436,10 +462,9 @@ class AMQPChannel extends AbstractChannel
         $source,
         $routing_key = '',
         $nowait = false,
-        $arguments = null,
+        $arguments = array(),
         $ticket = null
     ) {
-        $arguments = $this->getArguments($arguments);
         $ticket = $this->getTicket($ticket);
 
         list($class_id, $method_id, $args) = $this->protocolWriter->exchangeBind(
@@ -459,7 +484,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('exchange.bind_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -479,11 +504,17 @@ class AMQPChannel extends AbstractChannel
      * @param bool $nowait
      * @param array $arguments
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
-    public function exchange_unbind($destination, $source, $routing_key = '', $nowait = false, $arguments = null, $ticket = null)
-    {
-        $arguments = $this->getArguments($arguments);
+    public function exchange_unbind(
+        $destination,
+        $source,
+        $routing_key = '',
+        $nowait = false,
+        $arguments = array(),
+        $ticket = null
+    ) {
         $ticket = $this->getTicket($ticket);
 
         list($class_id, $method_id, $args) = $this->protocolWriter->exchangeUnbind(
@@ -499,7 +530,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('exchange.unbind_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -520,11 +551,17 @@ class AMQPChannel extends AbstractChannel
      * @param bool $nowait
      * @param array $arguments
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
-    public function queue_bind($queue, $exchange, $routing_key = '', $nowait = false, $arguments = null, $ticket = null)
-    {
-        $arguments = $this->getArguments($arguments);
+    public function queue_bind(
+        $queue,
+        $exchange,
+        $routing_key = '',
+        $nowait = false,
+        $arguments = array(),
+        $ticket = null
+    ) {
         $ticket = $this->getTicket($ticket);
 
         list($class_id, $method_id, $args) = $this->protocolWriter->queueBind(
@@ -544,7 +581,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('queue.bind_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -564,11 +601,16 @@ class AMQPChannel extends AbstractChannel
      * @param string $routing_key
      * @param array $arguments
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
-    public function queue_unbind($queue, $exchange, $routing_key = '', $arguments = null, $ticket = null)
-    {
-        $arguments = $this->getArguments($arguments);
+    public function queue_unbind(
+        $queue,
+        $exchange,
+        $routing_key = '',
+        $arguments = array(),
+        $ticket = null
+    ) {
         $ticket = $this->getTicket($ticket);
 
         list($class_id, $method_id, $args) = $this->protocolWriter->queueUnbind(
@@ -583,7 +625,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('queue.unbind_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -605,6 +647,7 @@ class AMQPChannel extends AbstractChannel
      * @param bool $nowait
      * @param array $arguments
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
     public function queue_declare(
@@ -614,10 +657,9 @@ class AMQPChannel extends AbstractChannel
         $exclusive = false,
         $auto_delete = true,
         $nowait = false,
-        $arguments = null,
+        $arguments = array(),
         $ticket = null
     ) {
-        $arguments = $this->getArguments($arguments);
         $ticket = $this->getTicket($ticket);
 
         list($class_id, $method_id, $args) = $this->protocolWriter->queueDeclare(
@@ -639,7 +681,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('queue.declare_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -665,6 +707,7 @@ class AMQPChannel extends AbstractChannel
      * @param bool $if_empty
      * @param bool $nowait
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
     public function queue_delete($queue = '', $if_unused = false, $if_empty = false, $nowait = false, $ticket = null)
@@ -687,7 +730,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('queue.delete_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -707,6 +750,7 @@ class AMQPChannel extends AbstractChannel
      * @param string $queue
      * @param bool $nowait
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|null
      */
     public function queue_purge($queue = '', $nowait = false, $ticket = null)
@@ -722,7 +766,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('queue.purge_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -858,6 +902,7 @@ class AMQPChannel extends AbstractChannel
      * @param string $consumer_tag
      * @param bool $nowait
      * @param bool $noreturn
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function basic_cancel($consumer_tag, $nowait = false, $noreturn = false)
@@ -872,7 +917,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('basic.cancel_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -898,6 +943,14 @@ class AMQPChannel extends AbstractChannel
         return $consumerTag;
     }
 
+    /**
+     * @return bool
+     */
+    public function is_consuming()
+    {
+        return !empty($this->callbacks);
+    }
+
     /**
      * Starts a queue consumer
      *
@@ -907,9 +960,10 @@ class AMQPChannel extends AbstractChannel
      * @param bool $no_ack
      * @param bool $exclusive
      * @param bool $nowait
-     * @param callback|null $callback
+     * @param callable|null $callback
      * @param int|null $ticket
      * @param array $arguments
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed|string
      */
     public function basic_consume(
@@ -940,7 +994,7 @@ class AMQPChannel extends AbstractChannel
         if (false === $nowait) {
             $consumer_tag = $this->wait(array(
                 $this->waitHelper->get_wait('basic.consume_ok')
-            ));
+            ), false, $this->channel_rpc_timeout);
         }
 
         $this->callbacks[$consumer_tag] = $callback;
@@ -993,6 +1047,7 @@ class AMQPChannel extends AbstractChannel
      * @param string $queue
      * @param bool $no_ack
      * @param int $ticket
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function basic_get($queue = '', $no_ack = false, $ticket = null)
@@ -1005,7 +1060,7 @@ class AMQPChannel extends AbstractChannel
         return $this->wait(array(
             $this->waitHelper->get_wait('basic.get_ok'),
             $this->waitHelper->get_wait('basic.get_empty')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -1104,15 +1159,20 @@ class AMQPChannel extends AbstractChannel
         $pkt = new AMQPWriter();
         $pkt->write($this->pre_publish($exchange, $routing_key, $mandatory, $immediate, $ticket));
 
-        $this->connection->send_content(
-            $this->channel_id,
-            60,
-            0,
-            mb_strlen($msg->body, 'ASCII'),
-            $msg->serialize_properties(),
-            $msg->body,
-            $pkt
-        );
+        try {
+            $this->connection->send_content(
+                $this->channel_id,
+                60,
+                0,
+                mb_strlen($msg->body, 'ASCII'),
+                $msg->serialize_properties(),
+                $msg->body,
+                $pkt
+            );
+        } catch (AMQPConnectionClosedException $e) {
+            $this->do_close();
+            throw $e;
+        }
 
         if ($this->next_delivery_tag > 0) {
             $this->published_messages[$this->next_delivery_tag] = $msg;
@@ -1191,6 +1251,7 @@ class AMQPChannel extends AbstractChannel
      * @param int $prefetch_size
      * @param int $prefetch_count
      * @param bool $a_global
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function basic_qos($prefetch_size, $prefetch_count, $a_global)
@@ -1205,7 +1266,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('basic.qos_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -1220,6 +1281,7 @@ class AMQPChannel extends AbstractChannel
      * Redelivers unacknowledged messages
      *
      * @param bool $requeue
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function basic_recover($requeue = false)
@@ -1229,7 +1291,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('basic.recover_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -1282,6 +1344,7 @@ class AMQPChannel extends AbstractChannel
     }
 
     /**
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function tx_commit()
@@ -1290,7 +1353,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('tx.commit_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -1304,6 +1367,7 @@ class AMQPChannel extends AbstractChannel
     /**
      * Rollbacks the current transaction
      *
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function tx_rollback()
@@ -1312,7 +1376,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('tx.rollback_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -1329,6 +1393,7 @@ class AMQPChannel extends AbstractChannel
      * Beware that only non-transactional channels may be put into confirm mode and vice versa
      *
      * @param bool $nowait
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return null
      */
     public function confirm_select($nowait = false)
@@ -1341,7 +1406,9 @@ class AMQPChannel extends AbstractChannel
             return null;
         }
 
-        $this->wait(array($this->waitHelper->get_wait('confirm.select_ok')));
+        $this->wait(array(
+            $this->waitHelper->get_wait('confirm.select_ok')
+        ), false, $this->channel_rpc_timeout);
         $this->next_delivery_tag = 1;
     }
 
@@ -1401,6 +1468,7 @@ class AMQPChannel extends AbstractChannel
     /**
      * Selects standard transaction mode
      *
+     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException if the specified operation timeout was exceeded
      * @return mixed
      */
     public function tx_select()
@@ -1409,7 +1477,7 @@ class AMQPChannel extends AbstractChannel
 
         return $this->wait(array(
             $this->waitHelper->get_wait('tx.select_ok')
-        ));
+        ), false, $this->channel_rpc_timeout);
     }
 
     /**
@@ -1426,6 +1494,10 @@ class AMQPChannel extends AbstractChannel
      */
     protected function getArguments($arguments)
     {
+        @trigger_error(sprintf(
+            'Method "%s" is deprecated, please use an array as a default argument instead',
+            __METHOD__
+        ), E_USER_DEPRECATED);
         return (null === $arguments) ? array() : $arguments;
     }
 
