@@ -20,6 +20,7 @@ class AMQPStreamConnection extends AbstractConnection
      * @param null $context
      * @param bool $keepalive
      * @param int $heartbeat
+     * @param float $channel_rpc_timeout
      */
     public function __construct(
         $host,
@@ -32,11 +33,17 @@ class AMQPStreamConnection extends AbstractConnection
         $login_response = null,
         $locale = 'en_US',
         $connection_timeout = 3.0,
-        $read_write_timeout = 3.0,
+        $read_write_timeout = 130.0,
         $context = null,
         $keepalive = false,
-        $heartbeat = 0
+        $heartbeat = 60,
+        $channel_rpc_timeout = 0.0,
+        $ssl_protocol = null
     ) {
+        if ($channel_rpc_timeout > $read_write_timeout) {
+            throw new \InvalidArgumentException('channel RPC timeout must not be greater than I/O read-write timeout');
+        }
+
         $io = new StreamIO(
             $host,
             $port,
@@ -44,7 +51,8 @@ class AMQPStreamConnection extends AbstractConnection
             $read_write_timeout,
             $context,
             $keepalive,
-            $heartbeat
+            $heartbeat,
+            $ssl_protocol
         );
 
         parent::__construct(
@@ -57,7 +65,8 @@ class AMQPStreamConnection extends AbstractConnection
             $locale,
             $io,
             $heartbeat,
-            $connection_timeout
+            $connection_timeout,
+            $channel_rpc_timeout
         );
 
         // save the params for the use of __clone, this will overwrite the parent
@@ -76,13 +85,13 @@ class AMQPStreamConnection extends AbstractConnection
         $connection_timeout = isset($options['connection_timeout']) ?
                                     $options['connection_timeout'] : 3.0;
         $read_write_timeout = isset($options['read_write_timeout']) ?
-                                    $options['read_write_timeout'] : 3.0;
+                                    $options['read_write_timeout'] : 130.0;
         $context = isset($options['context']) ?
                          $options['context'] : null;
         $keepalive = isset($options['keepalive']) ?
                            $options['keepalive'] : false;
         $heartbeat = isset($options['heartbeat']) ?
-                           $options['heartbeat'] : 0;
+                           $options['heartbeat'] : 60;
         return new static($host,
                           $port,
                           $user,
