@@ -3,66 +3,22 @@ namespace PhpAmqpLib\Wire\IO;
 
 use PhpAmqpLib\Exception\AMQPConnectionClosedException;
 use PhpAmqpLib\Exception\AMQPDataReadException;
-use PhpAmqpLib\Exception\AMQPHeartbeatMissedException;
 use PhpAmqpLib\Exception\AMQPIOException;
 use PhpAmqpLib\Exception\AMQPRuntimeException;
 use PhpAmqpLib\Exception\AMQPTimeoutException;
 use PhpAmqpLib\Helper\MiscHelper;
-use PhpAmqpLib\Wire\AMQPWriter;
 
 class StreamIO extends AbstractIO
 {
     /** @var string */
     protected $protocol;
 
-    /** @var string */
-    protected $host;
-
-    /** @var int */
-    protected $port;
-
-    /** @var float */
-    protected $connection_timeout;
-
-    /** @var float */
-    protected $read_write_timeout;
-
     /** @var resource */
     protected $context;
 
-    /** @var bool */
-    protected $keepalive;
-
-    /** @var int */
-    protected $heartbeat;
-
-    /** @var float */
-    protected $last_read;
-
-    /** @var float */
-    protected $last_write;
-
-    /** @var array */
-    protected $last_error;
-
-    /** @var int */
-    private $initial_heartbeat;
-
     /** @var resource */
     private $sock;
 
-    /** @var bool */
-    private $canDispatchPcntlSignal;
-
-    /** @var string */
-    private static $ERRNO_EQUALS_EAGAIN;
-
-    /** @var string */
-    private static $ERRNO_EQUALS_EWOULDBLOCK;
-
-    /** @var string */
-    private static $ERRNO_EQUALS_EINTR;
-
     /**
      * @param string $host
      * @param int $port
@@ -76,61 +32,55 @@ class StreamIO extends AbstractIO
         $host,
         $port,
         $connection_timeout,
-        $read_write_timeout,
+        $read_write_timeout = 130.0,
         $context = null,
         $keepalive = false,
-        $heartbeat = 0
+        $heartbeat = 60,
+        $ssl_protocol = null
     ) {
+        // TODO FUTURE change comparison to <=
+        // php-amqplib/php-amqplib#648, php-amqplib/php-amqplib#666
         if ($heartbeat !== 0 && ($read_write_timeout < ($heartbeat * 2))) {
             throw new \InvalidArgumentException('read_write_timeout must be at least 2x the heartbeat');
         }
 
-        self::$ERRNO_EQUALS_EAGAIN = 'errno=' . SOCKET_EAGAIN;
-        self::$ERRNO_EQUALS_EWOULDBLOCK = 'errno=' . SOCKET_EWOULDBLOCK;
-        self::$ERRNO_EQUALS_EINTR = 'errno=' . SOCKET_EINTR;
-
         $this->protocol = 'tcp';
         $this->host = $host;
         $this->port = $port;
         $this->connection_timeout = $connection_timeout;
-        $this->read_write_timeout = $read_write_timeout;
+        $this->read_timeout = $read_write_timeout;
+        $this->write_timeout = $read_write_timeout;
         $this->context = $context;
         $this->keepalive = $keepalive;
         $this->heartbeat = $heartbeat;
         $this->initial_heartbeat = $heartbeat;
         $this->canDispatchPcntlSignal = $this->isPcntlSignalEnabled();
 
-        if (is_null($this->context)) {
-            // tcp_nodelay was added in 7.1.0
-            if (PHP_VERSION_ID >= 70100) {
-                $this->context = stream_context_create(array(
-                    "socket" => array(
-                        "tcp_nodelay" => true
-                    )
-                ));
+        if (!is_resource($this->context) || get_resource_type($this->context) !== 'stream-context') {
+            $this->context = stream_context_create();
+        }
+
+        // tcp_nodelay was added in 7.1.0
+        if (PHP_VERSION_ID >= 70100) {
+            stream_context_set_option($this->context, 'socket', 'tcp_nodelay', true);
+        }
+
+        $options = stream_context_get_options($this->context);
+        if (!empty($options['ssl'])) {
+            if (isset($ssl_protocol)) {
+                $this->protocol = $ssl_protocol;
             } else {
-                $this->context = stream_context_create();
+                $this->protocol = 'ssl';
             }
-        } else {
-            $this->protocol = 'ssl';
+        }
+
+        if (!defined('SOCKET_EAGAIN')) {
+            define('SOCKET_EAGAIN', SOCKET_EWOULDBLOCK);
         }
     }
 
     /**
-     * @return bool
-     */
-    private function isPcntlSignalEnabled()
-    {
-        return extension_loaded('pcntl')
-            && function_exists('pcntl_signal_dispatch')
-            && (defined('AMQP_WITHOUT_SIGNALS') ? !AMQP_WITHOUT_SIGNALS : true);
-    }
-
-    /**
-     * Sets up the stream connection
-     *
-     * @throws \PhpAmqpLib\Exception\AMQPRuntimeException
-     * @throws \Exception
+     * @inheritdoc
      */
     public function connect()
     {
@@ -156,12 +106,11 @@ class StreamIO extends AbstractIO
             );
             $this->cleanup_error_handler();
         } catch (\ErrorException $e) {
-            restore_error_handler();
-            throw $e;
+            throw new AMQPIOException($e->getMessage());
         }
 
         if (false === $this->sock) {
-            throw new AMQPRuntimeException(
+            throw new AMQPIOException(
                 sprintf(
                     'Error Connecting to server(%s): %s ',
                     $errno,
@@ -172,7 +121,7 @@ class StreamIO extends AbstractIO
         }
 
         if (false === stream_socket_get_name($this->sock, true)) {
-            throw new AMQPRuntimeException(
+            throw new AMQPIOException(
                 sprintf(
                     'Connection refused: %s ',
                     $remote
@@ -180,7 +129,7 @@ class StreamIO extends AbstractIO
             );
         }
 
-        list($sec, $uSec) = MiscHelper::splitSecondsMicroseconds($this->read_write_timeout);
+        list($sec, $uSec) = MiscHelper::splitSecondsMicroseconds(max($this->read_timeout, $this->write_timeout));
         if (!stream_set_timeout($this->sock, $sec, $uSec)) {
             throw new AMQPIOException('Timeout could not be set');
         }
@@ -193,36 +142,23 @@ class StreamIO extends AbstractIO
                 stream_set_read_buffer($this->sock, 0);
             }
         } else {
-            stream_set_blocking($this->sock, 1);
+            stream_set_blocking($this->sock, true);
         }
 
         if ($this->keepalive) {
             $this->enable_keepalive();
         }
+        $this->heartbeat = $this->initial_heartbeat;
     }
 
     /**
-     * Reconnects the socket
-     */
-    public function reconnect()
-    {
-        $this->close();
-        $this->connect();
-    }
-
-    /**
-     * @param int $len
-     * @throws \ErrorException
-     * @throws \PhpAmqpLib\Exception\AMQPIOException
-     * @throws \PhpAmqpLib\Exception\AMQPDataReadException
-     * @return mixed|string
+     * @inheritdoc
      */
     public function read($len)
     {
         $this->check_heartbeat();
 
-        list($timeout_sec, $timeout_uSec) =
-            MiscHelper::splitSecondsMicroseconds($this->read_write_timeout);
+        list($timeout_sec, $timeout_uSec) = MiscHelper::splitSecondsMicroseconds($this->read_timeout);
 
         $read_start = microtime(true);
         $read = 0;
@@ -230,6 +166,7 @@ class StreamIO extends AbstractIO
 
         while ($read < $len) {
             if (!is_resource($this->sock) || feof($this->sock)) {
+                $this->close();
                 throw new AMQPConnectionClosedException('Broken pipe or closed connection');
             }
 
@@ -238,8 +175,7 @@ class StreamIO extends AbstractIO
                 $buffer = fread($this->sock, ($len - $read));
                 $this->cleanup_error_handler();
             } catch (\ErrorException $e) {
-                restore_error_handler();
-                throw $e;
+                throw new AMQPDataReadException($e->getMessage(), $e->getCode(), $e);
             }
 
             if ($buffer === false) {
@@ -248,15 +184,12 @@ class StreamIO extends AbstractIO
 
             if ($buffer === '') {
                 $read_now = microtime(true);
-                $t_read = round($read_now - $read_start);
-                if ($t_read > $this->read_write_timeout) {
+                $t_read = $read_now - $read_start;
+                if ($t_read > $this->read_timeout) {
                     throw new AMQPTimeoutException('Too many read attempts detected in StreamIO');
                 }
                 $this->select($timeout_sec, $timeout_uSec);
-                if ($this->canDispatchPcntlSignal) {
-                    pcntl_signal_dispatch();
-                }
-                $this->check_heartbeat();
+
                 continue;
             }
 
@@ -282,22 +215,20 @@ class StreamIO extends AbstractIO
     }
 
     /**
-     * @param string $data
-     * @return mixed|void
-     * @throws \PhpAmqpLib\Exception\AMQPRuntimeException
-     * @throws \PhpAmqpLib\Exception\AMQPTimeoutException
+     * @inheritdoc
      */
     public function write($data)
     {
         $written = 0;
         $len = mb_strlen($data, 'ASCII');
+        $write_start = microtime(true);
 
         while ($written < $len) {
-
             if (!is_resource($this->sock)) {
                 throw new AMQPConnectionClosedException('Broken pipe or closed connection');
             }
 
+            $result = false;
             $this->set_error_handler();
             // OpenSSL's C library function SSL_write() can balk on buffers > 8192
             // bytes in length, so we're limiting the write size here. On both TLS
@@ -307,123 +238,74 @@ class StreamIO extends AbstractIO
             // September 2002:
             // http://comments.gmane.org/gmane.comp.encryption.openssl.user/4361
             try {
-                $buffer = fwrite($this->sock, mb_substr($data, $written, 8192, 'ASCII'), 8192);
+                $buffer = mb_substr($data, $written, self::BUFFER_SIZE, 'ASCII');
+                $result = fwrite($this->sock, $buffer);
                 $this->cleanup_error_handler();
             } catch (\ErrorException $e) {
-                restore_error_handler();
-                throw new AMQPRuntimeException($e->getMessage());
+                $code = $this->last_error['errno'];
+                switch ($code) {
+                    case SOCKET_EPIPE:
+                    case SOCKET_ENETDOWN:
+                    case SOCKET_ENETUNREACH:
+                    case SOCKET_ENETRESET:
+                    case SOCKET_ECONNABORTED:
+                    case SOCKET_ECONNRESET:
+                    case SOCKET_ECONNREFUSED:
+                    case SOCKET_ETIMEDOUT:
+                        $this->close();
+                        throw new AMQPConnectionClosedException(socket_strerror($code), $code, $e);
+                    default:
+                        throw new AMQPRuntimeException($e->getMessage(), $code, $e);
+                }
             }
 
-            if ($buffer === false) {
+            if ($result === false) {
                 throw new AMQPRuntimeException('Error sending data');
             }
 
-            if ($buffer === 0 && feof($this->sock)) {
-                throw new AMQPConnectionClosedException('Broken pipe or closed connection');
-            }
-
             if ($this->timed_out()) {
-                throw new AMQPTimeoutException('Error sending data. Socket connection timed out');
+                throw AMQPTimeoutException::writeTimeout($this->write_timeout);
             }
 
-            $written += $buffer;
+            $now = microtime(true);
+            if ($result > 0) {
+                $this->last_write = $write_start = $now;
+                $written += $result;
+            } else {
+                if (feof($this->sock)) {
+                    $this->close();
+                    throw new AMQPConnectionClosedException('Broken pipe or closed connection');
+                }
+                if (($now - $write_start) > $this->write_timeout) {
+                    throw AMQPTimeoutException::writeTimeout($this->write_timeout);
+                }
+                // check stream and prevent from high CPU usage
+                $this->select_write();
+            }
         }
-
-        $this->last_write = microtime(true);
     }
 
     /**
-     * Internal error handler to deal with stream and socket errors that need to be ignored
-     *
-     * @param  int $errno
-     * @param  string $errstr
-     * @param  string $errfile
-     * @param  int $errline
-     * @param  array $errcontext
-     * @return null
-     * @throws \ErrorException
+     * @inheritdoc
      */
     public function error_handler($errno, $errstr, $errfile, $errline, $errcontext = null)
     {
-        // fwrite notice that the stream isn't ready - EAGAIN or EWOULDBLOCK
-        if (strpos($errstr, self::$ERRNO_EQUALS_EAGAIN) !== false || strpos($errstr, self::$ERRNO_EQUALS_EWOULDBLOCK) !== false) {
-             // it's allowed to retry
-            return null;
+        $code = $this->extract_error_code($errstr);
+        switch ($code) {
+            // fwrite notice that the stream isn't ready - EAGAIN or EWOULDBLOCK
+            case SOCKET_EAGAIN:
+            case SOCKET_EWOULDBLOCK:
+            // stream_select warning that it has been interrupted by a signal - EINTR
+            case SOCKET_EINTR:
+                return;
         }
 
-        // stream_select warning that it has been interrupted by a signal - EINTR
-        if (strpos($errstr, self::$ERRNO_EQUALS_EINTR) !== false) {
-             // it's allowed while processing signals
-            return null;
-        }
-
-        // throwing an exception in an error handler will halt execution
-        //   set the last error and continue
-        $this->last_error = compact('errno', 'errstr', 'errfile', 'errline', 'errcontext');
-    }
-
-    /**
-     * Begin tracking errors and set the error handler
-     */
-    protected function set_error_handler()
-    {
-        $this->last_error = null;
-        set_error_handler(array($this, 'error_handler'));
-    }
-
-    /**
-     * throws an ErrorException if an error was handled
-     */
-    protected function cleanup_error_handler()
-    {
-        if ($this->last_error !== null) {
-            throw new \ErrorException($this->last_error['errstr'], 0, $this->last_error['errno'], $this->last_error['errfile'], $this->last_error['errline']);
-        }
-
-        // no error was caught
-        restore_error_handler();
-    }
-
-    /**
-     * Heartbeat logic: check connection health here
-     * @throws \PhpAmqpLib\Exception\AMQPRuntimeException
-     */
-    public function check_heartbeat()
-    {
-        // ignore unless heartbeat interval is set
-        if ($this->heartbeat !== 0 && $this->last_read && $this->last_write) {
-            $t = microtime(true);
-            $t_read = round($t - $this->last_read);
-            $t_write = round($t - $this->last_write);
-
-            // server has gone away
-            if (($this->heartbeat * 2) < $t_read) {
-                $this->close();
-                throw new AMQPHeartbeatMissedException("Missed server heartbeat");
-            }
-
-            // time for client to send a heartbeat
-            if (($this->heartbeat / 2) < $t_write) {
-                $this->write_heartbeat();
-            }
-        }
-    }
-
-    /**
-     * Sends a heartbeat message
-     */
-    protected function write_heartbeat()
-    {
-        $pkt = new AMQPWriter();
-        $pkt->write_octet(8);
-        $pkt->write_short(0);
-        $pkt->write_long(0);
-        $pkt->write_octet(0xCE);
-        $this->write($pkt->getvalue());
+        parent::error_handler($code > 0 ? $code : $errno, $errstr, $errfile, $errline, $errcontext);
     }
 
     public function close()
     {
+        $this->disableHeartbeat();
         if (is_resource($this->sock)) {
             fclose($this->sock);
         }
@@ -433,51 +315,34 @@ class StreamIO extends AbstractIO
     }
 
     /**
-     * @return resource
+     * @inheritdoc
      */
-    public function get_socket()
+    public function getSocket()
     {
         return $this->sock;
     }
 
     /**
-     * @return resource
+     * @inheritdoc
      */
-    public function getSocket()
+    protected function do_select($sec, $usec)
     {
-        return $this->get_socket();
-    }
-
-    /**
-     * @param int $sec
-     * @param int $usec
-     * @return int|mixed
-     * @throws \ErrorException
-     * @throws \PhpAmqpLib\Exception\AMQPRuntimeException
-     */
-    public function select($sec, $usec)
-    {
-        $this->check_heartbeat();
-
         $read = array($this->sock);
         $write = null;
         $except = null;
-        $result = false;
 
-        if (defined('HHVM_VERSION')) {
-            $usec = is_int($usec) ? $usec : 0;
-        }
+        return stream_select($read, $write, $except, $sec, $usec);
+    }
 
-        $this->set_error_handler();
-        try {
-            $result = stream_select($read, $write, $except, $sec, $usec);
-            $this->cleanup_error_handler();
-        } catch (\ErrorException $e) {
-            restore_error_handler();
-            throw $e;
-        }
+    /**
+     * @return int|bool
+     */
+    protected function select_write()
+    {
+        $read = $except = null;
+        $write = array($this->sock);
 
-        return $result;
+        return stream_select($read, $write, $except, 0, 100000);
     }
 
     /**
@@ -496,6 +361,14 @@ class StreamIO extends AbstractIO
      */
     protected function enable_keepalive()
     {
+        if ($this->protocol === 'ssl') {
+            throw new AMQPIOException('Can not enable keepalive: ssl connection does not support keepalive (#70939)');
+        }
+
+        if ($this->protocol === 'tls') {
+            throw new AMQPIOException('Can not enable keepalive: tls connection does not support keepalive (#70939)');
+        }
+
         if (!function_exists('socket_import_stream')) {
             throw new AMQPIOException('Can not enable keepalive: function socket_import_stream does not exist');
         }
@@ -509,22 +382,22 @@ class StreamIO extends AbstractIO
     }
 
     /**
-     * @return $this
+     * @param string $message
+     * @return int
      */
-    public function disableHeartbeat()
+    protected function extract_error_code($message)
     {
-        $this->heartbeat = 0;
+        if (0 === strpos($message, 'stream_select():')) {
+            $pattern = '/\s+\[(\d+)\]:\s+/';
+        } else {
+            $pattern = '/\s+errno=(\d+)\s+/';
+        }
+        $matches = array();
+        $result = preg_match($pattern, $message, $matches);
+        if ($result > 0) {
+            return (int)$matches[1];
+        }
 
-        return $this;
-    }
-
-    /**
-     * @return $this
-     */
-    public function reenableHeartbeat()
-    {
-        $this->heartbeat = $this->initial_heartbeat;
-
-        return $this;
+        return 0;
     }
 }
